import sys, threading, time, Ice, NetRepository, Node
from util import Singleton
from util import MultipleReaderWriter

class NotReachable(Exception):
	pass

class New:
	__metaclass__ = Singleton.Cl

	def _init(self, p = 10020):
		self._nRep = NetRepository.New()

		self._retries = 4
		self._timeoutSec = 2.0
		self._timeoutMSec = int(self._timeoutSec * 1000.0)

		#stuff for communicators and adapters
		self._ftProt = 'tcp'
		self._ftPortS = str(p)
		self._ftPortC = str(p + 1)
		self._psProt = 'tcp'
		self._psPort = str(p + 2)

		#management for ip changes
		self._lock = MultipleReaderWriter.New()

		self._ftIceId2WrapperS = {}
		self._ftID2WrapperCProxyString = {}
		self._psIceId2WrapperS = {}
		self._psID2WrapperCProxyString = {}

		#management for object factories
		self._ftWrapperObjectFactories = {}
		self._psWrapperObjectFactories = {}

		#register callbacks and start ice
		self.start()
		if self._nRep.thread:
			self._nRep.addNsCb(self.shutdown)
			self._nRep.addCfgCb(self.start)

	def shutdown(self, netDict, news, losts):
		self._lock.acquireForWriting()
		sys.stderr.write('shutdown IN\n')

		#unregister wrapperS
		for iceId in self._ftIceId2WrapperS:
			self._ftAdapterS.remove(iceId)
		for iceId in self._psIceId2WrapperS:
			self._psAdapterS.remove(iceId)
		sys.stderr.write('shutdown A\n')
		#unregister object factories
		for iceId in self._ftWrapperObjectFactories:
			self._ftComS.removeObjectFactory(iceId)
			self._ftComC.removeObjectFactory(iceId)
		for iceId in self._psWrapperObjectFactories:
			self._psComS.removeObjectFactory(iceId)
			self._psComC.removeObjectFactory(iceId)
		sys.stderr.write('shutdown B\n')

		#shutdown communicators
		self._ftComS.shutdown()
		self._ftComS.destroy()
		self._ftComC.shutdown()
		self._ftComC.destroy()
		self._psCom.shutdown()
		self._psCom.destroy()

		#will be done by start
		#self._lock.releaseFromWriting()
		sys.stderr.write('shutdown OUT\n')

	def _changeProxy(self, ID):
		wrapper, proxyStrA, proxyStrB = self._ftID2WrapperCProxyString[ID]

		locator = wrapper.getLocator()
		cnt = locator.count('/')
		if cnt == 1:
			rID = locator[1 : ]
		else:
			rID = locator[1 : locator.find('/', 1)]

		try:
			proxyStr = proxyStrA + self._nRep.getMicroNet(rID).ip[0] + proxyStrB
			proxy = self._ftComC.stringToProxy(proxyStr)
			type = wrapper.getType()
			exec('import ICE' + type.split('.')[0])
			wrapper.proxy = eval('ICE' + type + 'Prx.uncheckedCast(proxy)')
		except:
			pass

	def start(self):
		#was done by shutdone
		#self._lock.acquireForWriting()
		sys.stderr.write('start IN\n')

		#configure ice
		props = Ice.getDefaultProperties(sys.argv)
		props.setProperty('Ice.RetryIntervals', '-1')
		props.setProperty('Ice.Override.Timeout', str(self._timeoutMSec))
		props.setProperty('Ice.ThreadPool.Client.Size', '10')
		props.setProperty('Ice.ThreadPool.Client.SizeMax', '100')
		props.setProperty('Ice.ThreadPool.Server.Size', '10')
		props.setProperty('Ice.ThreadPool.Server.SizeMax', '100')
# 		props.setProperty('Ice.GC.Interval', '5')

		#debuging
#		props.setProperty('Ice.Trace.Protocol', '1')
#		props.setProperty('Ice.Trace.Network', '2')
#		props.setProperty('Ice.ThreadPool.Client.SizeWarn', '90')
#		props.setProperty('Ice.ThreadPool.Server.SizeWarn', '90')

		if self._nRep.ip:
			ip = self._nRep.ip
		else:
			#happens if net wasn't stable at startup
			ip = '127.0.0.1'

		if 'getwindowsversion' in sys.__dict__:
			ip = '0.0.0.0'

		sys.stderr.write('IP ' + ip + '\n')

		#create communicators
		self._ftComS = Ice.initialize()
		self._ftComC = Ice.initialize()
		self._psCom = Ice.initialize()

		#register object factories
		for iceId in self._ftWrapperObjectFactories:
			objF = self._ftWrapperObjectFactories[iceId]
			self._ftComS.addObjectFactory(objF, iceId)
			self._ftComC.addObjectFactory(objF, iceId)
		for iceId in self._psWrapperObjectFactories:
			objF = self._psWrapperObjectFactories[iceId]
			self._psComS.addObjectFactory(objF, iceId)
			self._psComC.addObjectFactory(objF, iceId)

		#create adapters
		self._ftAdapterS = self._ftComS.createObjectAdapterWithEndpoints(
			self._nRep.ID, 
			self._ftProt + \
			' -h ' + ip + \
			' -p ' + self._ftPortS
		)
		self._ftAdapterC = self._ftComC.createObjectAdapterWithEndpoints(
			self._nRep.ID, 
			self._ftProt + \
			' -h ' + ip + \
			' -p ' + self._ftPortC
		)
		self._psAdapterS = self._psCom.createObjectAdapterWithEndpoints( \
			self._nRep.ID, 
			self._psProt + \
			' -h ' + ip + \
			' -p ' + self._psPort
		)
		self._psAdapterC = self._psAdapterS

		#fill adapters
		for iceId in self._ftIceId2WrapperS:
			self._ftAdapterS.add(self._ftIceId2WrapperS[iceId], iceId)
		for ID in self._ftID2WrapperCProxyString:
			self._changeProxy(ID)
		for iceId in self._psIceId2WrapperS:
			self._psAdapterS.add(self._psIceId2WrapperS[iceId], iceId)
		for ID in self._psID2WrapperCProxyString:
			self._changeProxy(ID)

		#active adapters
		self._ftAdapterS.activate()
		self._ftAdapterC.activate()
		self._psAdapterS.activate()

		#may be aquired by self.shutdown
		if self._lock.isLockedForWriting():
			self._lock.releaseFromWriting()
		sys.stderr.write('start OUT\n')

	def addObjectFactory(self, objF, iceId, tp = 'ft'):
		self._lock.acquireForWriting()

		try:
			if tp == 'ft':
				self._ftWrapperObjectFactories[iceId] = objF
				self._ftComS.addObjectFactory(objF, iceId)
				self._ftComC.addObjectFactory(objF, iceId)
			else:
				self._psWrapperObjectFactories[iceId] = objF
				self._psComS.addObjectFactory(objF, iceId)
				self._psComC.addObjectFactory(objF, iceId)
		except:
			pass

		self._lock.releaseFromWriting()

	def _createIceId(self, obj):
		if issubclass(obj.__class__, Node.New):
			module = 'Node'
		else:
			module = obj.__module__
		iceId = module + '.' + obj.__class__.__name__ + '!' + obj.getLocator().replace('/', '#') + '!' + obj.getID()
		return Ice.stringToIdentity(iceId)

	def regWrapperS(self, wrapper, tp = 'ft'):
		self._lock.acquireForWriting()

		wrapper.iceId = self._createIceId(wrapper._obj)
		if tp == 'ft':
			self._ftAdapterS.add(wrapper, wrapper.iceId)
			self._ftIceId2WrapperS[wrapper.iceId] = wrapper
		else:
			self._psAdapterS.add(wrapper, wrapper.iceId)
			self._psIceId2WrapperS[wrapper.iceId] = wrapper

		self._lock.releaseFromWriting()

	def unregWrapperS(self, wrapper, tp = 'ft'):
		self._lock.acquireForWriting()

		if tp == 'ft':
			del self._ftIceId2WrapperS[wrapper.iceId]
			self._ftAdapterS.remove(wrapper.iceId)
		else:
			del self._psIceId2WrapperS[wrapper.iceId]
			self._psAdapterS.remove(wrapper.iceId)

		self._lock.releaseFromWriting()

	def regWrapperC(self, wrapper, proxyStr, tp = 'ft'):
		self._lock.acquireForWriting()

		#register wrapper
		a = proxyStr.find(' -h ') + 4
		b = proxyStr.find(' ', a)
		proxyStrA = proxyStr[ : a]
		proxyStrB = proxyStr[b : ]
		if tp == 'ft':
			self._ftID2WrapperCProxyString[wrapper.getLocator()] = wrapper, proxyStrA, proxyStrB
		else:
			self._psID2WrapperCProxyString[wrapper.getLocator()] = wrapper, proxyStrA, proxyStrB

		#create proxy
		if tp == 'ft':
			proxy = self._ftComC.stringToProxy(proxyStr)
		else:
			proxy = self._psCom.stringToProxy(proxyStr)

		type = wrapper.getType()
		exec('import ICE' + type.split('.')[0])
		wrapper.proxy = eval('ICE' + type + 'Prx.uncheckedCast(proxy)')

		self._lock.releaseFromWriting()

	def unregWrapperC(self, wrapper, tp = 'ft'):
		self._lock.acquireForWriting()

		if tp == 'ft':
			self._ftID2WrapperCProxyString[wrapper.getLocator()]
		else:
			self._psID2WrapperCProxyString[wrapper.getLocator()]

		self._lock.releaseFromWriting()

	def invoke(self, wrapper, fName, *params):
		#don't acquire lock because connection errors
		#lead to sleep and retries
		for i in range(self._retries):
			try:
				self._lock.acquireForReading()
				sys.stderr.write('invoke IN\n')

				sys.stderr.write(str(i) + ' ' + str(wrapper.proxy) + '\n')

				ret  = eval('wrapper.proxy.' + fName + '(*params)')

				self._lock.releaseFromReading()
				sys.stderr.write('invoke OUT 1\n')

				return ret
			except Ice.UserException, e:
				self._lock.releaseFromReading()
				sys.stderr.write('invoke OUT 2\n')
				raise e
			except Ice.TimeoutException:
				pass
			except:
				time.sleep(self._timeoutSec)

			self._lock.releaseFromReading()
			sys.stderr.write('invoke OUT 3\n')

		raise NotReachable()

	def createWrapperCFromProxyStr(self, proxyStr):
		s = 0
		if proxyStr[0] == '"':
			s = 1
		module, cl = proxyStr[s : proxyStr.find('!')].split('.')
		exec('import ' + module + 'C')
		return eval(module + 'C.' + cl + '(proxyStr)')

	def createProxyStringFromObject(self, obj):
		self._lock.acquireForWriting()

		iceId = self._createIceId(obj)
		a = self._ftAdapterS.find(iceId)
		b = self._psAdapterS.find(iceId)

		if not a and not b:
			self._lock.releaseFromWriting()
			raise ICEIceManager.NotNetable()
		if a:
			ret = str(self._ftAdapterS.createProxy(iceId))
		else:
			ret = str(self._psAdapterS.createProxy(iceId))

		self._lock.releaseFromWriting()

		return ret
